
import './utils/cc_vec_extension';

import {game_events} from "./game_events";
import executionOrder = cc._decorator.executionOrder;
import menu = cc._decorator.menu;
const {ccclass, property, executeInEditMode} = cc._decorator;

@ccclass()
@executeInEditMode()
@executionOrder(-1)
@menu('layer/rhombic layout')
export class RhombicLayout extends cc.Component{
    @property
    get edge_cells(){return this._edge_cells}
    set edge_cells(value){
        this._edge_cells = value;
        this.update_total_cells();
        this.update_y_axis_cell_count();
        this.update_relate_properties();
    }

    @property()
    get cell_size(){return this._cell_size;}
    set cell_size(value){
        this._cell_size.set(value);
        this.update_half_cell_size();
        this.update_relate_properties();
    }

    @property({tooltip:'space between tow cell'})
    get cell_space(){return this._cell_space;}
    set cell_space(value){
        this._cell_space.set(value);
        this.update_relate_properties();
    }


    @property
    get total_cells(){return this._total_cells;}

    @property({type:cc.Graphics})
    get graph(){return this.graphics}

    get y_axis_cell_count(){return this._y_axis_cell_count;}
    get size(){return this._size;}

    /**
     * position the nodes with cell_centers
     * @param {cc.Node[]} nodes
     */
    layout(nodes:cc.Node[]){
        if(nodes.length !== this.total_cells){
            cc.error(`cell_count not equal nodes count.cell count: ${this.total_cells}, node count:${nodes.length}`);
            return;
        }
        for(let i=0; i<this.total_cells; ++i){
            nodes[i].setPosition(this.cell_centers[i]);
        }
    }

    get_col_count(row){
        return this.edge_cells - Math.abs(row - this.edge_cells + 1);
    }

    private update_total_cells(){
        this._total_cells = this._edge_cells * this._edge_cells;
        cc.director.emit(game_events.total_cells_changed, this.total_cells);
    }
    /**
     * related with size_cells, cell_size, cell_space.
     */
    private update_size(){
        let count = this.edge_cells;
        let space_count = count-1;
        let w = count * this.cell_size.x + space_count * this.cell_space.x;
        let h = count * this.cell_size.y + space_count * this.cell_space.y;
        this._size.x = w;
        this._size.y = h;
        this.node.setContentSize(w,h);
    }

    private update_cells_center(){
        //cc.log('update cell centers');
        let index = 0;
        let offset_rows = (this.y_axis_cell_count - 1) / 2.0;
        this.cell_centers = [];
        for(let row = offset_rows; row >= -offset_rows; --row){
            //cc.log('row:', row);
            let y =  (this.cell_size.y + this.cell_space.y) * row / 2;
            let cols = this.get_col_begin(row);
            let offset_cols = (cols-1) / 2;
            //cc.log('cols:',cols);
            for(let col=-offset_cols; col <= offset_cols; ++col){
                //cc.log('row:',row, 'col:',col);
                let x = (this.cell_size.x + this.cell_space.x) * col;
                this.cell_centers[index++] = new cc.Vec2(x,y);
            }
        }
        //cc.log('after update cell centers:', this.cell_centers);
    }

    private get_col_begin(row){
        return this.edge_cells - Math.abs(row);
    }

    private update_y_axis_cell_count(){
        this._y_axis_cell_count = 2 * this.edge_cells - 1;
    }

    private init_graphics(){
        let debug_node = this.node.get_or_add_child('_debug');
        debug_node.zIndex = 999;
        this.graphics = debug_node.get_or_add_component(cc.Graphics);
    }
    private draw_gizmo(){
        this.graphics.clear();
        this.draw_border();
        this.graphics.stroke();
        this.draw_cells_border();
        this.graphics.stroke();
        this.draw_cell_centers();
        this.graphics.stroke();
        this.draw_cells_border();
    }
    private draw_border(){
        this.graphics.strokeColor = cc.Color.RED;
        this.graphics.polyline(this.border_points);
    }
    private draw_cell_centers(){
        this.graphics.strokeColor = cc.Color.GREEN;
        for(let i=0; i< this.total_cells; ++i){
            //cc.log('cell center:', this.cell_centers[i]);
            this.graphics.draw_cross(this.cell_centers[i], 10, cc.Color.GREEN)
        }
        this.graphics.stroke();
    }
    private draw_cells_border(){
        this.graphics.strokeColor = cc.Color.GRAY;
        for(let points of this.cells_border_points){
            this.graphics.polyline(points);
        }
    }

    /**
     * properties relate with edge_cells, cell_size, cell_space,
     */
    private update_relate_properties(){
        this.update_size();
        this.update_border_points();
        this.update_cells_center();
        this.update_cells_border_points();
    }
    /**
     * call after update_size;
     */
    private update_border_points(){
        let half_w = this.size.x / 2;
        let half_h = this.size.y / 2;
        this.border_points[0] = new cc.Vec2(0, half_h);
        this.border_points[1] = new cc.Vec2(half_w, 0);
        this.border_points[2] = new cc.Vec2(0, -half_h);
        this.border_points[3] = new cc.Vec2(-half_w, 0);
    }

    private update_cells_border_points(){
        this.cells_border_points = [];
        for (let center of this.cell_centers) {
            let points = [];
            for(let i=0;i<4;++i){
                points[i] = RhombicLayout.get_corner_point(center, this.half_cell_size, i);
            }
            //cc.log('points:', points);
            this.cells_border_points.push(points);
        }
    }


    private static get_corner_point(center:cc.Vec2,half_size:cc.Vec2,index:number){
        switch (index){
            case 0:
                return center.add(new cc.Vec2(-half_size.x,0));
            case 1:
                return center.add(new cc.Vec2(0, -half_size.y));
            case 2:
                return center.add(new cc.Vec2(half_size.x, 0));
            case 3:
                return center.add(new cc.Vec2(0, half_size.y));
        }
    }
    private init_half_cell_size(){
        this.half_cell_size = new cc.Vec2(this.cell_size.x / 2, this.cell_size.y / 2);
    }
    private update_half_cell_size(){
        this.half_cell_size.set_xy(this.cell_size.x/ 2, this.cell_size.y / 2);
    }


    onLoad(){
        this.init_half_cell_size();
        this.update_y_axis_cell_count();
        this.update_size();
        this.init_graphics();
        this.update_border_points();
        this.update_cells_center();
        this.update_cells_border_points();
    }
    start(){
        cc.director.emit(game_events.total_cells_changed, this.total_cells);
    }
    onRestore(){
        this.onLoad();
    }
    update(dt){
        this.draw_gizmo();
    }
    onDisable(){
        this.graphics.clear();
    }


    @property()
    private _edge_cells:number = 4;
    private _cell_size:cc.Vec2 = new cc.Vec2(220,103);
    @property(cc.Vec2)
    private _cell_space:cc.Vec2 = new cc.Vec2(5,80);
    private half_cell_size:cc.Vec2 = new cc.Vec2(0,0);
    @property()
    private _total_cells:number = 16;
    private _size:cc.Vec2 = new cc.Vec2();
    private cell_centers:cc.Vec2[] = [];
    private _y_axis_cell_count:number = 0;
    private graphics:cc.Graphics;
    private border_points:cc.Vec2[] = [];
    private cells_border_points:cc.Vec2[][];
}


